import{k as _,m as r,s as d,K as m,n as c,v as u,w as f,x as v,M as g,N as b,O as h,F as G,G as C,H as k}from"./index.ec2afa6c.js";function p(n){let s,i;const o=n[3].default,t=m(o,n,n[2],null);return{c(){s=c("div"),t&&t.c(),u(s,"id",n[0]),u(s,"class","svelte-10ogue4"),f(s,"hidden",!n[1])},m(e,l){v(e,s,l),t&&t.m(s,null),i=!0},p(e,[l]){t&&t.p&&(!i||l&4)&&g(t,o,e,e[2],i?h(o,e[2],l,null):b(e[2]),null),(!i||l&1)&&u(s,"id",e[0]),l&2&&f(s,"hidden",!e[1])},i(e){i||(G(t,e),i=!0)},o(e){C(t,e),i=!1},d(e){e&&k(s),t&&t.d(e)}}}function q(n,s,i){let{$$slots:o={},$$scope:t}=s,{elem_id:e=""}=s,{visible:l=!0}=s;return n.$$set=a=>{"elem_id"in a&&i(0,e=a.elem_id),"visible"in a&&i(1,l=a.visible),"$$scope"in a&&i(2,t=a.$$scope)},[e,l,t,o]}class w extends _{constructor(s){super(),r(this,s,q,p,d,{elem_id:0,visible:1})}}var H=w;const K=["static"];export{H as Component,K as modes};
